﻿using System;
using System.Net;
using System.Net.Sockets;
using System.Threading.Tasks;

namespace EonaCat.Network
{
    public class SocketTcpClient
    {
        public event Action<RemoteInfo> OnConnect;
        public event Action<RemoteInfo> OnReceive;
        public event Action<RemoteInfo> OnDisconnect;
        public event Action<RemoteInfo> OnSend;
        public event Action<Exception, string> OnError;

        private Socket socket;

        public Task ConnectAsync(string ipAddress, int port)
        {
            if (!IPAddress.TryParse(ipAddress, out IPAddress ip))
            {
                throw new Exception("Invalid ipAddress given");
            }

            return CreateSocketTcpClientAsync(ip, port);
        }

        private async Task CreateSocketTcpClientAsync(IPAddress ipAddress, int port)
        {
            IsIP6 = ipAddress.AddressFamily == AddressFamily.InterNetworkV6;
            socket = new Socket(IsIP6 ? AddressFamily.InterNetworkV6 : AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            try
            {
                await socket.ConnectAsync(ipAddress, port).ConfigureAwait(false);
                OnConnect?.Invoke(new RemoteInfo { IsTcp = true, IsIpv6 = IsIP6 });
                await StartReceivingAsync();
            }
            catch (SocketException ex)
            {
                OnError?.Invoke(ex, $"SocketException: {ex.Message}");
                Disconnect();
            }
        }

        public bool IsIP6 { get; private set; }

        private async Task StartReceivingAsync()
        {
            byte[] buffer = new byte[1024];
            while (socket.Connected)
            {
                try
                {
                    int received = await socket.ReceiveAsync(new ArraySegment<byte>(buffer), SocketFlags.None).ConfigureAwait(false);
                    if (received > 0)
                    {
                        byte[] data = buffer.AsSpan(0, received).ToArray();
                        OnReceive?.Invoke(new RemoteInfo
                        {
                            IsTcp = true,
                            Data = data,
                            EndPoint = socket.RemoteEndPoint,
                            IsIpv6 = IsIP6
                        });
                    }
                    else
                    {
                        break;
                    }
                }
                catch (SocketException ex)
                {
                    OnError?.Invoke(ex, $"SocketException: {ex.Message}");
                    break;
                }
            }
            OnDisconnect?.Invoke(new RemoteInfo { IsTcp = true, EndPoint = socket.RemoteEndPoint, IsIpv6 = IsIP6 });
        }

        public async Task SendAsync(byte[] data)
        {
            await socket.SendAsync(new ArraySegment<byte>(data), SocketFlags.None).ConfigureAwait(false);
            OnSend?.Invoke(new RemoteInfo { EndPoint = socket.RemoteEndPoint, Data = data, IsIpv6 = IsIP6 });
        }

        public void Disconnect()
        {
            socket.Close();
        }
    }
}
