﻿using System.Net.Sockets;
using System.Net;
using System;
using System.Threading.Tasks;
using System.Threading;

namespace EonaCat.Network
{
    public class SocketTcpServer
    {
        public event Action<RemoteInfo> OnConnect;
        public event Action<RemoteInfo> OnReceive;
        public event Action<RemoteInfo> OnSend;
        public event Action<RemoteInfo> OnDisconnect;
        public event Action<Exception, string> OnError;

        private TcpListener listener;

        public SocketTcpServer(IPAddress ipAddress, int port)
        {
            listener = new TcpListener(ipAddress, port);
        }

        public SocketTcpServer(string ipAddress, int port)
        {
            IPAddress address = IPAddress.Parse(ipAddress);
            listener = new TcpListener(address, port);
        }

        public async Task StartAsync(CancellationToken cancellationToken = default)
        {
            listener.Start();
            while (!cancellationToken.IsCancellationRequested)
            {
                try
                {
                    Socket socket = await listener.AcceptSocketAsync().ConfigureAwait(false);
                    OnConnect?.Invoke(new RemoteInfo
                    {
                        Socket = socket,
                        IsTcp = true,
                        IsIpv6 = socket.RemoteEndPoint.AddressFamily == AddressFamily.InterNetworkV6,
                        EndPoint = socket.RemoteEndPoint
                    });
                    await StartReceivingAsync(socket, cancellationToken).ConfigureAwait(false);
                }
                catch (SocketException ex)
                {
                    OnError?.Invoke(ex, $"SocketException: {ex.Message}");
                }
            }
        }

        private async Task StartReceivingAsync(Socket socket, CancellationToken cancellationToken)
        {
            byte[] buffer = new byte[1024];
            while (!cancellationToken.IsCancellationRequested)
            {
                try
                {
                    int received = await socket.ReceiveAsync(new Memory<byte>(buffer), SocketFlags.None, cancellationToken)
                        .ConfigureAwait(false);
                    if (received > 0)
                    {
                        OnReceive?.Invoke(new RemoteInfo
                        {
                            Socket = socket,
                            IsTcp = true,
                            IsIpv6 = socket.RemoteEndPoint.AddressFamily == AddressFamily.InterNetworkV6,
                            EndPoint = socket.RemoteEndPoint,
                            Data = new Memory<byte>(buffer, 0, received).ToArray()
                        });
                    }
                    else
                    {
                        break;
                    }
                }
                catch (SocketException ex)
                {
                    OnError?.Invoke(ex, $"SocketException: {ex.Message}");
                    break;
                }
            }
        }

        public async Task SendTo(Socket socket, byte[] data)
        {
            await socket.SendAsync(new ArraySegment<byte>(data), SocketFlags.None).ConfigureAwait(false);
            OnSend?.Invoke(new RemoteInfo { Socket = socket, IsTcp = true, EndPoint = socket.RemoteEndPoint, IsIpv6 = socket.RemoteEndPoint.AddressFamily == AddressFamily.InterNetworkV6, Data = data });
        }
    }
}