﻿using System.Net.Sockets;
using System.Net;
using System.Threading.Tasks;
using System;
using System.Threading;

namespace EonaCat.Network
{
    public class SocketUdpClient
    {
        public event Action<EndPoint> OnConnect;
        public event Action<byte[]> OnReceive;
        public event Action<EndPoint> OnDisconnect;
        public event Action<EndPoint, byte[]> OnSend;
        public event Action<Exception, string> OnError;

        private UdpClient udpClient;

        public SocketUdpClient(IPAddress ipAddress, int port, CancellationToken cancellationToken = default)
        {
            CreateUdpClient(ipAddress, port, cancellationToken);
        }

        private void CreateUdpClient(IPAddress ipAddress, int port, CancellationToken cancellationToken = default)
        {
            IsIP6 = ipAddress.AddressFamily == AddressFamily.InterNetworkV6;
            udpClient = new UdpClient(ipAddress.AddressFamily);

            if (IsIP6)
            {
                udpClient.Client.SetSocketOption(SocketOptionLevel.IPv6, SocketOptionName.IPv6Only, false);
            }

            udpClient.Client.Bind(new IPEndPoint(IsIP6 ? IPAddress.IPv6Any : IPAddress.Any, 0));

            if (IsMulticastGroupEnabled)
            {
                udpClient.JoinMulticastGroup(ipAddress);
            }

            _ = StartReceivingAsync(cancellationToken);
        }

        public bool IsMulticastGroupEnabled { get; set; }

        public bool IsIP6 { get; private set; }

        public SocketUdpClient(string ipAddress, int port)
        {
            if (!IPAddress.TryParse(ipAddress, out IPAddress ip))
            {
                throw new Exception("Invalid ipAddress given");
            }

            CreateUdpClient(ip, port);
        }

        private async Task StartReceivingAsync(CancellationToken cancellationToken = default)
        {
            while (!cancellationToken.IsCancellationRequested)
            {
                try
                {
                    UdpReceiveResult result = await udpClient.ReceiveAsync().ConfigureAwait(false);
                    OnReceive?.Invoke(result.Buffer);
                }
                catch (SocketException ex)
                {
                    OnError?.Invoke(ex, $"SocketException: {ex.Message}");
                    break;
                }
            }
            OnDisconnect?.Invoke(udpClient.Client.RemoteEndPoint);
        }

        public async Task SendTo(EndPoint endPoint, byte[] data)
        {
            if (endPoint is IPEndPoint ipEndPoint)
            {
                await udpClient.SendAsync(data, data.Length, ipEndPoint).ConfigureAwait(false);
                OnSend?.Invoke(endPoint, data);
            }
        }

        public void Disconnect()
        {
            udpClient.Close();
        }
    }
}
